from __future__ import annotations

from lonboard._constants import EXTENSION_NAME


def is_native_geoarrow(extension_type_name: bytes | None) -> bool:
    """Return True if this GeoArrow column has a "native" coordinate representation

    This will return false for WKB and WKT-serialized arrays.
    """
    return extension_type_name not in {
        EXTENSION_NAME.WKB,
        EXTENSION_NAME.OGC_WKB,
        EXTENSION_NAME.WKT,
    }



"""Handle GeoArrow tables with WKB-encoded geometry"""

import json
from typing import Dict, List

import numpy as np
from arro3.core import Table

from lonboard._constants import EXTENSION_NAME, OGC_84
from lonboard._geoarrow.crs import get_field_crs
from lonboard._geoarrow.extension_types import construct_geometry_array
from lonboard._geoarrow.utils import is_native_geoarrow
from lonboard._utils import get_geometry_column_index


def parse_serialized_table(table: Table) -> List[Table]:
    """Parse a table with a serialized WKB/WKT column into GeoArrow-native geometries.

    If no columns are WKB/WKT-encoded, returns the input. Note that WKB columns must be
    tagged with an extension name of `geoarrow.wkb` or `ogc.wkb`. WKT columns must be
    tagged with an extension name of `geoarrow.wkt`.

    Returns one table per GeoArrow geometry type. E.g. if the input has points, lines,
    and polygons, then it returns three tables. Point/MultiPoint,
    LineString/MultiLineString, Polygon/MultiPolygon are each combined into a single
    table type.
    """
    table = parse_geoparquet_table(table)
    field_idx = get_geometry_column_index(table.schema)

    # For non-geometry table, just return as-is
    if field_idx is None:
        return [table]

    field = table.field(field_idx)
    column = table.column(field_idx)

    extension_type_name = field.metadata.get(b"ARROW:extension:name")

    # For native GeoArrow input, return table as-is
    if is_native_geoarrow(extension_type_name):
        return [table]

    import shapely
    from shapely import GeometryType

    # Handle WKB/WKT input
    crs = get_field_crs(field)
    if extension_type_name in {EXTENSION_NAME.WKB, EXTENSION_NAME.OGC_WKB}:
        shapely_arr = shapely.from_wkb(column)
    elif extension_type_name == EXTENSION_NAME.WKT:
        shapely_arr = shapely.from_wkt(column)
    else:
        raise ValueError(f"Unexpected GeoArrow extension name {extension_type_name}")

    type_ids = np.array(shapely.get_type_id(shapely_arr))
    unique_type_ids = set(np.unique(type_ids))

    if GeometryType.GEOMETRYCOLLECTION in unique_type_ids:
        raise ValueError("GeometryCollections not currently supported")

    if GeometryType.LINEARRING in unique_type_ids:
        raise ValueError("LinearRings not currently supported")

    point_indices = np.where(
        (type_ids == GeometryType.POINT) | (type_ids == GeometryType.MULTIPOINT)
    )[0]

    linestring_indices = np.where(
        (type_ids == GeometryType.LINESTRING)
        | (type_ids == GeometryType.MULTILINESTRING)
    )[0]

    polygon_indices = np.where(
        (type_ids == GeometryType.POLYGON) | (type_ids == GeometryType.MULTIPOLYGON)
    )[0]

    # Here we intentionally check geometries in a specific order.
    # Starting from polygons, then linestrings, then points,
    # so that the order of generated layers is polygon, then path then scatterplot.
    # This ensures that points are rendered on top and polygons on the bottom.
    parsed_tables: List[Table] = []
    for single_type_geometry_indices in (
        polygon_indices,
        linestring_indices,
        point_indices,
    ):
        if len(single_type_geometry_indices) == 0:
            continue

        single_type_geometry_field, single_type_geometry_arr = construct_geometry_array(
            shapely_arr[single_type_geometry_indices],
            crs=crs,
        )

        concatted_table = table.combine_chunks()
        batches = concatted_table.to_batches()
        assert len(batches) == 1

        assert single_type_geometry_indices.dtype == np.int64

        single_type_geometry_record_batch = (
            batches[0]
            .take(single_type_geometry_indices)
            .set_column(field_idx, single_type_geometry_field, single_type_geometry_arr)
        )
        parsed_tables.append(Table.from_batches([single_type_geometry_record_batch]))

    return parsed_tables


def parse_geoparquet_table(table: Table) -> Table:
    """Parse GeoParquet table metadata, assigning it to GeoArrow metadata"""
    # If a column already has geoarrow metadata, don't parse from GeoParquet metadata
    if get_geometry_column_index(table.schema) is not None:
        return table

    schema_metadata = table.schema.metadata or {}
    geo_metadata = schema_metadata.get(b"geo")
    if not geo_metadata:
        return table

    try:
        geo_metadata = json.loads(geo_metadata)
    except json.JSONDecodeError:
        return table

    primary_column = geo_metadata["primary_column"]
    column_meta = geo_metadata["columns"][primary_column]
    column_idx = [
        idx for idx, name in enumerate(table.column_names) if name == primary_column
    ]
    assert len(column_idx) == 1, f"Expected one column with name {primary_column}"
    column_idx = column_idx[0]
    if column_meta["encoding"] == "WKB":
        existing_field = table.schema.field(column_idx)
        existing_column = table.column(column_idx)
        crs_metadata = {"crs": column_meta.get("crs", OGC_84.to_json_dict())}
        metadata: Dict[bytes, bytes] = {
            b"ARROW:extension:name": EXTENSION_NAME.WKB,
            b"ARROW:extension:metadata": json.dumps(crs_metadata).encode(),
        }
        new_field = existing_field.with_metadata(metadata)
        table = table.set_column(column_idx, new_field, existing_column)

    return table



import json
from typing import Optional

from arro3.core import Field


# Note: According to the spec, if the metadata key exists, its value should never be
# `null` or an empty dict, but we still check for those to be safe
def get_field_crs(field: Field) -> Optional[dict]:
    extension_metadata_value = field.metadata.get(b"ARROW:extension:metadata")
    if not extension_metadata_value:
        return None

    extension_metadata = json.loads(extension_metadata_value)
    return extension_metadata.get("crs")



import sys
from typing import Any, Dict, Union

from lonboard.basemap import CartoBasemap

if sys.version_info >= (3, 12):
    from typing import TypedDict
else:
    from typing_extensions import TypedDict


class MapKwargs(TypedDict, total=False):
    _height: int
    basemap_style: Union[str, CartoBasemap]
    parameters: Dict[str, Any]
    picking_radius: int
    show_tooltip: bool
    use_device_pixels: Union[int, float, bool]
    view_state: Dict[str, Any]



from __future__ import annotations

import json
from enum import Enum
from typing import TYPE_CHECKING, Dict, Optional, Sequence, Tuple

import numpy as np
from arro3.core import Array, DataType, Field, fixed_size_list_array, list_array

if TYPE_CHECKING:
    from numpy.typing import NDArray


class CoordinateDimension(str, Enum):
    XY = "xy"
    XYZ = "xyz"
    XYM = "xym"
    XYZM = "xyzm"


def coord_storage_type(*, interleaved: bool, dims: CoordinateDimension) -> DataType:
    """Generate the storage type of a geoarrow coordinate array

    Args:
        interleaved: Whether coordinates should be interleaved or separated
        dims: The number of dimensions
    """
    if interleaved:
        return DataType.list(Field(dims, DataType.float64()), len(dims))

    else:
        if dims == CoordinateDimension.XY:
            return DataType.struct(
                [
                    Field("x", DataType.float64()),
                    Field("y", DataType.float64()),
                ]
            )
        if dims == CoordinateDimension.XYZ:
            return DataType.struct(
                [
                    Field("x", DataType.float64()),
                    Field("y", DataType.float64()),
                    Field("z", DataType.float64()),
                ]
            )
        if dims == CoordinateDimension.XYM:
            return DataType.struct(
                [
                    Field("x", DataType.float64()),
                    Field("y", DataType.float64()),
                    Field("m", DataType.float64()),
                ]
            )
        if dims == CoordinateDimension.XYZM:
            return DataType.struct(
                [
                    Field("x", DataType.float64()),
                    Field("y", DataType.float64()),
                    Field("z", DataType.float64()),
                    Field("m", DataType.float64()),
                ]
            )


def linestring_storage_type(
    *, interleaved: bool, dims: CoordinateDimension, large_list: bool = False
) -> DataType:
    """Generate the storage type of a geoarrow.linestring array

    Args:
        interleaved: Whether coordinates should be interleaved or separated
        dims: The number of dimensions
        large_list: Whether to use a large list with int64 offsets for the inner type
    """
    vertices_type = coord_storage_type(interleaved=interleaved, dims=dims)
    if large_list:
        return DataType.large_list(Field("vertices", vertices_type))
    else:
        return DataType.list(Field("vertices", vertices_type))


def polygon_storage_type(
    *, interleaved: bool, dims: CoordinateDimension, large_list: bool = False
) -> DataType:
    """Generate the storage type of a geoarrow.polygon array

    Args:
        interleaved: Whether coordinates should be interleaved or separated
        dims: The number of dimensions
        large_list: Whether to use a large list with int64 offsets for the inner type
    """
    rings_type = linestring_storage_type(
        large_list=large_list, interleaved=interleaved, dims=dims
    )
    if large_list:
        return DataType.large_list(Field("rings", rings_type))
    else:
        return DataType.list(Field("rings", rings_type))


def multipoint_storage_type(
    *, interleaved: bool, dims: CoordinateDimension, large_list: bool = False
) -> DataType:
    """Generate the storage type of a geoarrow.multipoint array

    Args:
        interleaved: Whether coordinates should be interleaved or separated
        dims: The number of dimensions
        large_list: Whether to use a large list with int64 offsets for the inner type
    """
    points_type = coord_storage_type(interleaved=interleaved, dims=dims)
    if large_list:
        return DataType.large_list(Field("points", points_type))
    else:
        return DataType.list(Field("points", points_type))


def multilinestring_storage_type(
    *, interleaved: bool, dims: CoordinateDimension, large_list: bool = False
) -> DataType:
    """Generate the storage type of a geoarrow.multilinestring array

    Args:
        interleaved: Whether coordinates should be interleaved or separated
        dims: The number of dimensions
        large_list: Whether to use a large list with int64 offsets for the inner type
    """
    linestrings_type = linestring_storage_type(
        large_list=large_list, interleaved=interleaved, dims=dims
    )
    if large_list:
        return DataType.large_list(Field("linestrings", linestrings_type))
    else:
        return DataType.list(Field("linestrings", linestrings_type))


def multipolygon_storage_type(
    *, interleaved: bool, dims: CoordinateDimension, large_list: bool = False
) -> DataType:
    """Generate the storage type of a geoarrow.multipolygon array

    Args:
        interleaved: Whether coordinates should be interleaved or separated
        dims: The number of dimensions
        large_list: Whether to use a large list with int64 offsets for the inner type
    """
    polygons_type = polygon_storage_type(
        large_list=large_list, interleaved=interleaved, dims=dims
    )
    if large_list:
        return DataType.large_list(Field("polygons", polygons_type))
    else:
        return DataType.list(Field("polygons", polygons_type))


def offsets_to_arrow(
    offsets: Tuple[NDArray[np.int64], ...],
) -> Sequence[Array]:
    # Shapely produces int64 offset arrays. We downcast those to int32 if possible
    if any(offset_arr[-1] >= np.iinfo(np.int32).max for offset_arr in offsets):
        return [Array(offset_arr) for offset_arr in offsets]

    return [Array(offset_arr.astype(np.int32)) for offset_arr in offsets]


def construct_geometry_array(
    shapely_arr: NDArray[np.object_],
    include_z: Optional[bool] = None,
    *,
    field_name: str = "geometry",
    crs: Optional[dict] = None,
) -> Tuple[Field, Array]:
    import shapely
    from shapely import GeometryType

    geom_type, coords, offsets = shapely.to_ragged_array(
        shapely_arr, include_z=include_z
    )
    offsets = offsets_to_arrow(offsets)

    if coords.shape[-1] == 2:
        dims = CoordinateDimension.XY
    elif coords.shape[-1] == 3:
        dims = CoordinateDimension.XYZ
    else:
        raise ValueError(f"Unexpected coords dimensions: {coords.shape}")

    extension_metadata: Dict[str, str] = {}
    if crs is not None:
        extension_metadata["ARROW:extension:metadata"] = json.dumps({"crs": crs})

    if geom_type == GeometryType.POINT:
        arrow_coords = fixed_size_list_array(coords.ravel("C"), len(dims))
        extension_metadata["ARROW:extension:name"] = "geoarrow.point"
        field = Field(
            field_name,
            arrow_coords.type,
            nullable=True,
            metadata=extension_metadata,
        )
        return field, arrow_coords

    elif geom_type == GeometryType.LINESTRING:
        assert len(offsets) == 1, "Expected one offsets array"
        (geom_offsets,) = offsets
        arrow_coords = fixed_size_list_array(coords.ravel("C"), len(dims))
        arrow_geoms = list_array(geom_offsets, arrow_coords)
        extension_metadata["ARROW:extension:name"] = "geoarrow.linestring"
        field = Field(
            field_name,
            arrow_geoms.type,
            nullable=True,
            metadata=extension_metadata,
        )
        return field, arrow_geoms

    elif geom_type == GeometryType.POLYGON:
        assert len(offsets) == 2, "Expected two offsets arrays"
        ring_offsets, geom_offsets = offsets
        arrow_coords = fixed_size_list_array(coords.ravel("C"), len(dims))
        arrow_rings = list_array(ring_offsets, arrow_coords)
        arrow_geoms = list_array(geom_offsets, arrow_rings)
        extension_metadata["ARROW:extension:name"] = "geoarrow.polygon"
        field = Field(
            field_name,
            arrow_geoms.type,
            nullable=True,
            metadata=extension_metadata,
        )
        return field, arrow_geoms

    elif geom_type == GeometryType.MULTIPOINT:
        assert len(offsets) == 1, "Expected one offsets array"
        (geom_offsets,) = offsets
        arrow_coords = fixed_size_list_array(coords.ravel("C"), len(dims))
        arrow_geoms = list_array(geom_offsets, arrow_coords)
        extension_metadata["ARROW:extension:name"] = "geoarrow.multipoint"
        field = Field(
            field_name,
            arrow_geoms.type,
            nullable=True,
            metadata=extension_metadata,
        )
        return field, arrow_geoms

    elif geom_type == GeometryType.MULTILINESTRING:
        assert len(offsets) == 2, "Expected two offsets arrays"
        ring_offsets, geom_offsets = offsets
        arrow_coords = fixed_size_list_array(coords.ravel("C"), len(dims))
        arrow_rings = list_array(ring_offsets, arrow_coords)
        arrow_geoms = list_array(geom_offsets, arrow_rings)
        extension_metadata["ARROW:extension:name"] = "geoarrow.multilinestring"
        field = Field(
            field_name,
            arrow_geoms.type,
            nullable=True,
            metadata=extension_metadata,
        )
        return field, arrow_geoms

    elif geom_type == GeometryType.MULTIPOLYGON:
        assert len(offsets) == 3, "Expected three offsets arrays"
        ring_offsets, polygon_offsets, geom_offsets = offsets
        arrow_coords = fixed_size_list_array(coords.ravel("C"), len(dims))
        arrow_rings = list_array(ring_offsets, arrow_coords)
        arrow_polygons = list_array(polygon_offsets, arrow_rings)
        arrow_geoms = list_array(geom_offsets, arrow_polygons)
        extension_metadata["ARROW:extension:name"] = "geoarrow.multipolygon"
        field = Field(
            field_name,
            arrow_geoms.type,
            nullable=True,
            metadata=extension_metadata,
        )
        return field, arrow_geoms

    else:
        raise ValueError(f"Unsupported type for geoarrow: {geom_type}")
